@@ -46,6 +46,8 @@ typedef struct _Listener Listener;
 
 struct _Listener {
   guint cnxn;
+  guint refcount : 24;
+  guint removed : 1; /* has been removed */
   gpointer listener_data;
   GFreeFunc destroy_notify;
 };
@@ -98,6 +100,9 @@ static gboolean ltable_get_data (LTable *ltable,
                                  guint cnxn_id,
                                  gpointer *listener_data_p,
                                  const gchar  **location_p);
+static void    ltable_remove_if  (LTable* ltable,
+                                  GConfListenersPredicate predicate,
+                                  gpointer user_data);
 
 #ifdef DEBUG_LISTENERS
 static void    ltable_spew(LTable* ltable);
@@ -107,8 +112,11 @@ static LTableEntry* ltable_entry_new(const gchar* name,
                                      const gchar* full_name);
 static void         ltable_entry_destroy(LTableEntry* entry);
 
-static Listener* listener_new(guint cnxn_id, gpointer listener_data, GFreeFunc destroy_notify);
-static void listener_destroy(Listener* l);
+static Listener* listener_new   (guint      cnxn_id,
+                                 gpointer   listener_data,
+                                 GFreeFunc  destroy_notify);
+static void      listener_unref (Listener  *l);
+static void      listener_ref   (Listener  *l);
 
 /*
  * Public API
@@ -156,10 +164,9 @@ gconf_listeners_remove  (GConfListeners* listeners,
 {
   LTable* lt = (LTable*)listeners;
 
-  g_return_if_fail(cnxn_id > 0);
-
-  /* This is a programmer error but we want to be robust even with
-     checks turned off */
+  /* Silently fail, since this can easily happen (e.g. adding the
+   * notify fails)
+   */
   if (cnxn_id == 0)
     return;
   
@@ -206,6 +213,16 @@ gconf_listeners_get_data (GConfListeners* listeners,
   return ltable_get_data (lt, cnxn_id, listener_data_p, location_p);
 }
 
+void
+gconf_listeners_remove_if (GConfListeners         *listeners,
+                           GConfListenersPredicate predicate,
+                           gpointer                user_data)
+{
+  LTable* lt = (LTable*)listeners;
+
+  ltable_remove_if (lt, predicate, user_data);
+}
+
 /*
  * LTable impl
  */
@@ -217,18 +234,30 @@ listener_new(guint cnxn_id, gpointer listener_data, GFreeFunc destroy_notify)
 
   l = g_new0(Listener, 1);
 
+  l->refcount = 1;
+  l->removed = FALSE;
   l->listener_data = listener_data;
-  l->cnxn = cnxn_id;
+  l->cnxn = cnxn_id;  
   l->destroy_notify = destroy_notify;
 
   return l;
 }
 
 static void      
-listener_destroy(Listener* l)
+listener_unref (Listener* l)
 {
-  (*l->destroy_notify)(l->listener_data);
-  g_free(l);
+  l->refcount -= 1;
+  if (l->refcount == 0)
+    {
+      (*l->destroy_notify)(l->listener_data);
+      g_free(l);
+    }
+}
+
+static void
+listener_ref (Listener  *l)
+{
+  l->refcount += 1;
 }
 
 static LTable* 
@@ -308,7 +337,7 @@ ltable_insert(LTable* lt, const gchar* where, Listener* l)
   const gchar* noroot_where = where + 1;
 
   g_return_if_fail(gconf_valid_key(where, NULL));
-
+  
   if (lt->tree == NULL)
     {
       lte = ltable_entry_new("/", "/");
@@ -390,8 +419,8 @@ ltable_insert(LTable* lt, const gchar* where, Listener* l)
   lt->active_listeners += 1;
 
 #ifdef DEBUG_LISTENERS
-  printf("Added %u at %s, spewing:\n",
-         l->cnxn, where);
+  g_print ("Added %u at %s, spewing:\n",
+	   l->cnxn, where);
   ltable_spew(lt);
 #endif
 }
@@ -403,8 +432,9 @@ ltable_remove(LTable* lt, guint cnxn)
   GList* tmp;
   GNode* node;
   guint index = CNXN_ID_INDEX(cnxn);
-  
+
   g_return_if_fail(index < lt->listeners->len);
+  
   if (index >= lt->listeners->len) /* robust even with checks off */
     return;
   
@@ -446,8 +476,9 @@ ltable_remove(LTable* lt, guint cnxn)
 
           lt->removed_indices = g_slist_prepend(lt->removed_indices,
                                                 GUINT_TO_POINTER(index));
-          
-          listener_destroy(l);
+
+          l->removed = TRUE;
+          listener_unref (l);
 
           break;
         }
@@ -468,31 +499,31 @@ ltable_remove(LTable* lt, guint cnxn)
   g_ptr_array_index(lt->listeners, index) = NULL;
 
   /* Remove from the tree if this node is now pointless */
-    {
-      GNode* cur = node;
-      while (cur != NULL)
-        {
-          GNode* parent = cur->parent;
-          lte = cur->data;
-          if (lte->listeners == NULL && cur->children == NULL)
-            {
-              if (cur == lt->tree)
-                lt->tree = NULL;
+  {
+    GNode* cur = node;
+    while (cur != NULL)
+      {
+        GNode* parent = cur->parent;
+        lte = cur->data;
+        if (lte->listeners == NULL && cur->children == NULL)
+          {
+            if (cur == lt->tree)
+              lt->tree = NULL;
               
-              ltable_entry_destroy(lte);
-              g_node_destroy(cur);
-            }
-          else
-            break; /* don't delete more parents since this node exists */
+            ltable_entry_destroy(lte);
+            g_node_destroy(cur);
+          }
+        else
+          break; /* don't delete more parents since this node exists */
           
-          cur = parent;
-        }
-    }
+        cur = parent;
+      }
+  }
 
   lt->active_listeners -= 1;
 
 #ifdef DEBUG_LISTENERS
-  printf("Removed %u, spewing:\n", cnxn);
+  g_print ("Removed %u, spewing:\n", cnxn);
   ltable_spew(lt);
 #endif
 }
@@ -507,15 +538,16 @@ destroy_func(GNode* node, gpointer data)
   while (tmp != NULL)
     {
       Listener* l = tmp->data;
-      
-      listener_destroy(l);
+
+      l->removed = TRUE;
+      listener_unref (l);
 
       tmp = g_list_next(tmp);
     }
-  g_list_free(lte->listeners);
+  g_list_free (lte->listeners);
   lte->listeners = NULL;
   
-  ltable_entry_destroy(lte);
+  ltable_entry_destroy (lte);
   
   return FALSE;
 }
@@ -551,7 +583,9 @@ notify_listener_list(GConfListeners* listeners,
     {
       Listener* l = tmp->data;
 
-      (*callback)(listeners, key, l->cnxn, l->listener_data, user_data);
+      /* don't notify listeners that were removed during the notify */
+      if (!l->removed)
+        (*callback)(listeners, key, l->cnxn, l->listener_data, user_data);
 
       tmp = g_list_next(tmp);
     }
@@ -565,7 +599,8 @@ ltable_notify(LTable* lt, const gchar* key,
   guint i;
   const gchar* noroot_key;
   GNode* cur;
-
+  GList *to_notify;
+  
   noroot_key = key + 1;
 
   g_return_if_fail(*key == '/');
@@ -573,13 +608,15 @@ ltable_notify(LTable* lt, const gchar* key,
 
   if (lt->tree == NULL)
     return; /* no one to notify */
+
+  /* we build a list "to_notify" to be safe against tree
+   * modifications during the notification.
+   */
   
   /* Notify "/" listeners */
-  notify_listener_list((GConfListeners*)lt,
-                       ((LTableEntry*)lt->tree->data)->listeners, 
-                       key, callback, user_data);
+  to_notify = g_list_copy (((LTableEntry*)lt->tree->data)->listeners);  
 
-  dirs = g_strsplit(noroot_key, "/", -1);
+  dirs = g_strsplit (noroot_key, "/", -1);
 
   cur = lt->tree;
   i = 0;
@@ -593,9 +630,9 @@ ltable_notify(LTable* lt, const gchar* key,
 
           if (strcmp(lte->name, dirs[i]) == 0)
             {
-              notify_listener_list((GConfListeners*)lt,
-                                   lte->listeners, key,
-                                   callback, user_data);
+              GList *copy = g_list_copy (lte->listeners);
+              to_notify = g_list_concat (to_notify, copy);
+
               break;
             }
 
@@ -611,6 +648,16 @@ ltable_notify(LTable* lt, const gchar* key,
     }
   
   g_strfreev(dirs);
+
+  g_list_foreach (to_notify, (GFunc) listener_ref, NULL);
+  
+  notify_listener_list ((GConfListeners*)lt,
+                        to_notify, key,
+                        callback, user_data);
+
+  g_list_foreach (to_notify, (GFunc) listener_unref, NULL);
+
+  g_list_free (to_notify);
 }
 
 struct NodeTraverseData
@@ -712,6 +759,72 @@ ltable_get_data (LTable *lt,
   return FALSE;
 }
 
+
+
+struct NodeRemoveData
+{
+  GConfListenersPredicate predicate;
+  gpointer user_data;
+  GSList *dead;
+};
+
+static gboolean
+node_remove_func (GNode   *node,
+                  gpointer user_data)
+{
+  struct NodeRemoveData *rd = user_data;
+  LTableEntry *lte = node->data;
+  GList *tmp;
+
+  tmp = lte->listeners;
+  while (tmp != NULL)
+    {
+      Listener *l = tmp->data;
+      
+      if ((* rd->predicate) (lte->full_name,
+                             l->cnxn,
+                             l->listener_data,
+                             rd->user_data))
+        rd->dead = g_slist_prepend (rd->dead, GINT_TO_POINTER (l->cnxn));
+      
+      tmp = g_list_next (tmp);
+    }
+
+  return FALSE; /* continue traversal */
+}
+
+static void
+ltable_remove_if  (LTable                 *ltable,
+                   GConfListenersPredicate predicate,
+                   gpointer                user_data)
+{
+  GSList *tmp;
+  struct NodeRemoveData rd;
+  rd.predicate = predicate;
+  rd.user_data = user_data;
+  rd.dead = NULL;
+  
+  if (ltable->tree == NULL)
+    return;
+  
+  g_node_traverse (ltable->tree,
+                   G_IN_ORDER,
+                   G_TRAVERSE_ALL,
+                   -1,
+                   node_remove_func,
+                   &rd);
+
+  tmp = rd.dead;
+  while (tmp != NULL)
+    {
+      ltable_remove (ltable, GPOINTER_TO_INT (tmp->data));
+
+      tmp = g_slist_next (tmp);
+    }
+
+  g_slist_free (rd.dead);
+}
+
 static LTableEntry* 
 ltable_entry_new(const gchar* name,
                  const gchar *full_name)
@@ -747,20 +860,20 @@ spew_func(GNode* node, gpointer data)
   spaces[g_node_depth(node)] = '\0';
 
   
-  printf(" %sSpewing node `%s' (%p): ", spaces, lte->name, node);
+  g_print (" %sSpewing node `%s' (%p): ", spaces, lte->name, node);
 
   tmp = lte->listeners;
   while (tmp != NULL)
     {
       Listener* l = tmp->data;
       
-      printf("  %slistener %u is here\n", spaces, (guint)l->cnxn);
+      g_print ("  %slistener %u is here\n", spaces, (guint)l->cnxn);
 
       tmp = g_list_next(tmp);
     }
 
   if (lte->listeners == NULL)
-    printf("\n");
+    g_print ("\n");
   
   return FALSE;
 }
@@ -769,13 +882,13 @@ static void
 ltable_spew(LTable* lt)
 {
   guint i;
-  printf("Flat table:\n");
+  g_print ("Flat table:\n");
   i = 0;
   while (i < lt->listeners->len)
     {
       GNode* node = g_ptr_array_index(lt->listeners, i);
       LTableEntry* lte = node ? node->data : NULL;
-      printf("%u `%s' %p\n", i, lte ? lte->name : "", node);
+      g_print ("%u `%s' %p\n", i, lte ? lte->name : "", node);
       
       ++i;
     }
